import argparse
import logging
import random
import threading
import time
import uuid

from .logging import setup_log


class Master:
    def __init__(self, max_threads=10):
        self.max_threads = max_threads
        self.threads = []

    def run(self, target, target_args):
        while True:
            while len(self.threads) < self.max_threads:
                thread = threading.Thread(target=target, args=target_args)
                logging.info('New thread is spawned: %s', thread.name)
                thread.start()
                self.threads.append(thread)
            threads = []
            for thread in self.threads:
                if thread.is_alive():
                    threads.append(thread)
                else:
                    logging.info('Thread is terminated: %s', thread.name)
            self.threads = threads


def test():
    x = random.randint(1, 30)
    n = uuid.uuid4()
    logging.info('Sleeping %s seconds: %s', x, n)
    time.sleep(x)
    logging.info('Sleeping is completed: %s', n)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', action='store_true', default=False, help='show debug information')
    args, _ = parser.parse_known_args()
    setup_log(level=logging.DEBUG if args.debug else logging.INFO)
    Master().run(test, ())



from .list import EnhancedList


class EnhancedDict:
    def __init__(self, d: dict):
        self.d = d.copy()

    def search(self, key, full=False, dive=True):
        """
        Search dict and find all items with certain keys
        :param key: key
        :param full: the entire dicts will be returned as results if set to true
        :param dive: continue searching even if key is already matched
        :return: list of items
        """
        result = []
        for k, v in self.d.items():
            if key == k:
                result.append(self.d if full else v)
            if dive:
                if isinstance(v, dict):
                    result += EnhancedDict(v).search(key, full, dive)
                elif isinstance(v, list):
                    result += EnhancedList(EnhancedList(v).format(lambda x: EnhancedDict(x).search(key, full, dive) if isinstance(x, dict) else None)).flatten()
        return result

    def merge(self, d: dict):
        """
        Merge another dict into this dict
        :param d: the other dict
        :return: the merged dict (self.d)
        """
        for k in d:
            if k in self.d and isinstance(self.d[k], dict) and isinstance(d[k], dict):
                self.d[k] = EnhancedDict(self.d[k]).merge(d[k])
            else:
                self.d[k] = d[k]
        return self.d

    def flatten(self):
        """
        Replace nested dict keys to underscore-connected keys
        :return: flattened dictionary
        """
        result = dict()
        for k, v in self.d.items():
            if isinstance(v, dict):
                for kk, vv in EnhancedDict(v).flatten().items():
                    result[k + '_' + kk] = vv
            else:
                result[k] = v
        return result

    def format(self, formatter, picker=None):
        """
        Format the values of the dict
        :param formatter: formatter method, e.g., return the string format of an int
        :param picker: picker method, e.g., return not none
        :return: formatted dictionary (self.d)
        """
        for k in self.d.keys():
            if isinstance(self.d[k], list):
                self.d[k] = EnhancedList(self.d[k]).format(
                    lambda x: EnhancedDict(x).format(formatter, picker) if isinstance(x, dict) else (
                        formatter(x) if picker is None or (picker is not None and picker(x)) else x
                    )
                )
            elif isinstance(self.d[k], dict):
                self.d[k] = EnhancedDict(self.d[k]).format(formatter, picker)
            elif picker is None or (picker is not None and picker(self.d[k])):
                self.d[k] = formatter(self.d[k])
        return self.d

    # def dict_remove_key(d, k):
    #     """
    #     Recursively remove a key from a dict
    #     :param d: the dictionary
    #     :param k: key which should be removed
    #     :return: formatted dictionary
    #     """
    #     if isinstance(d, dict):
    #         dd = dict()
    #         for key, value in d.items():
    #             if not key == k:
    #                 dd[key] = dict_remove_key(value, k)
    #         return dd
    #     elif isinstance(d, list):
    #         return [dict_remove_key(i, k) for i in d]
    #     else:
    #         return d
    #
    # def dict_remove_value(d, v):
    #     """
    #     Recursively remove keys with a certain value from a dict
    #     :param d: the dictionary
    #     :param v: value which should be removed
    #     :return: formatted dictionary
    #     """
    #     dd = dict()
    #     for key, value in d.items():
    #         if not value == v:
    #             if isinstance(value, dict):
    #                 dd[key] = dict_remove_value(value, v)
    #             elif isinstance(value, list):
    #                 dd[key] = [dict_remove_value(i, v) for i in value]
    #             else:
    #                 dd[key] = value
    #     return dd
    #
    # def dict_as_tuple_list(d, as_list=False):
    #     """
    #     Format a dict to a list of tuples
    #     :param d: the dictionary
    #     :param as_list: return a list of lists rather than a list of tuples
    #     :return: formatted dictionary list
    #     """
    #     dd = list()
    #     for k, v in d.items():
    #         dd.append([k, v] if as_list else (k, v))
    #     return dd


# class EnhancedDictList:
#     def __init__(self, m: list):
#         self.d = d
#
#     def search_by_key_value(self, key, value):
#         """
#         Search dictionary list by key and value
#         :param d: dictionary list
#         :param k: key
#         :param v: value
#         :return: the index of the first dictionary in the array with the specific key / value
#         """
#         for i in range(len(d)):
#             if d[i][k] == v:
#                 return i
#         return None
#
#     def dict_sort(d, k):
#         """
#         Sort a dictionary list by key
#         :param d: dictionary list
#         :param k: key
#         :return: sorted dictionary list
#         """
#         return sorted(d.copy(), key=lambda i: i[k])
#
#     def dict_top(d, k, n, reverse=False):
#         """
#         Return top n of a dictionary list sorted by key
#         :param d: dictionary list
#         :param k: key
#         :param n: top n
#         :param reverse: whether the value should be reversed
#         :return: top n of the sorted dictionary list
#         """
#         h = list()
#         for i in range(len(d)):
#             heappush(h, (-d[i][k] if reverse else d[i][k], i))
#         r = list()
#         while len(r) < n and len(h) > 0:
#             _, i = heappop(h)
#             r.append(d[i].copy())
#         return r
#
#     def tuple_search(t, i, v):
#         """
#         Search tuple array by index and value
#         :param t: tuple array
#         :param i: index of the value in each tuple
#         :param v: value
#         :return: the first tuple in the array with the specific index / value
#         """
#         for e in t:
#             if e[i] == v:
#                 return e
#         return None


if __name__ == "__main__":
    dict_a = {
        'a': 1,
        'b': {
            'a': 1,
            'b': {
                'a': 1,
                'b': [{
                    'a': 1,
                    'b': 2
                }, {
                    'a': 1,
                    'b': {
                        'a': 1,
                        'b': 2
                    }
                }]
            }
        }
    }
    dict_b = {
        'a': 1,
        'b': {
            'b': {
                'b': 2
            }
        }
    }
    print(dict_a)
    print(dict_b)
    print(EnhancedDict(dict_a).search('a'))
    print(EnhancedDict(dict_a).merge(dict_b))
    print(EnhancedDict(dict_b).flatten())
    print(EnhancedDict(dict_a).format(lambda x: x + 10, lambda x: x == 1))



from decimal import Decimal, InvalidOperation, ROUND_HALF_DOWN


class EnhancedDecimal:
    def __init__(self, n, default=None):
        if n is None:
            if isinstance(default, Decimal):
                self.n = default
            else:
                raise TypeError
        elif not isinstance(n, Decimal):
            try:
                self.n = Decimal(repr(n))
            except InvalidOperation:
                self.n = Decimal(n)
        else:
            self.n = n

    def round(self, r: Decimal, d=ROUND_HALF_DOWN):
        """
        Round the decimal number with a given unit
        :param r: the unit, e.g., 1, 0.5
        :param d: rounding direction, see decimal.ROUND_* for details
        :return: the rounded number
        """
        r = EnhancedDecimal(r)
        return (self.n / r.n).to_integral_exact(rounding=d) * r.n


if __name__ == "__main__":
    print(EnhancedDecimal('3.1415926').round(Decimal('0.1')))



class EnhancedList:
    def __init__(self, r: list):
        self.r = r.copy()

    def flatten(self):
        """
        Remove all sub-lists of the list
        :return: flattened list
        """
        result = []
        for i in self.r:
            if isinstance(i, list):
                result += EnhancedList(i).flatten()
            else:
                result.append(i)
        return result

    def pick(self, picker, reverse=False):
        """
        Pick values from the original list
        :param picker: picker method, e.g,: return true if value is not none
        :param reverse: return values that do not match the picker
        :return: picked list
        """
        result = []
        for i in self.r:
            if isinstance(i, list):
                candidates = EnhancedList(i).pick(picker, reverse)
                if candidates:
                    result.append(candidates)
            else:
                if picker(i) ^ reverse:
                    result.append(i)
        return result

    def format(self, formatter, picker=None):
        """
        Format the values of the list
        :param formatter: formatter method, e.g., return the string format of an int
        :param picker: picker method, e.g., return not none
        :return: formatted list (self.r)
        """
        for i in range(len(self.r)):
            if isinstance(self.r[i], list):
                self.r[i] = EnhancedList(self.r[i]).format(formatter, picker)
            elif picker is None or (picker is not None and picker(self.r[i])):
                self.r[i] = formatter(self.r[i])
        return self.r


if __name__ == "__main__":
    list_a = [1, [1, [2, 3], 2], 2]
    print(list_a)
    print(EnhancedList(list_a).flatten())
    print(EnhancedList(list_a).pick(lambda x: x == 1))
    print(EnhancedList(list_a).pick(lambda x: x == 1, reverse=True))
    print(EnhancedList(list_a).format(lambda x: x + 1, lambda x: isinstance(x, (int, float))))



import logging
import os
import sys
import time
from datetime import datetime

from django.utils.termcolors import colorize


class Formatter(logging.Formatter):
    color = {
        logging.DEBUG: 'cyan',
        logging.INFO: 'green',
        logging.WARNING: 'yellow',
        logging.ERROR: 'red',
        logging.CRITICAL: 'magenta',
    }

    def format(self, record):
        original = self._style._fmt
        self._style._fmt = colorize('[%(asctime)s] ', fg=self.color[record.levelno]) + '%(message)s'
        result = logging.Formatter.format(self, record)
        self._style._fmt = original
        return result


def setup_log(level=logging.INFO, path=None, tag: str = None):
    """
    Setup logging with color formats
    :param level: logging level via logging.INFO / logging.WARNING / logging.ERROR, etc.
    :param path: logs will be written to a file instead of stdout if provided
    :param tag: a prefix of the log file if path if provided
    :return: none
    """
    # Stream handler
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(Formatter())
    logging.root.addHandler(handler)
    # File handler
    if path is not None:
        if not os.path.exists(path):
            os.makedirs(path)
        handler = logging.FileHandler("{}/{}{}.log".format(path, tag + '-' if tag is not None else '', datetime.utcnow().timestamp()), encoding='utf-8')
        handler.setFormatter(logging.Formatter('[%(asctime)s] %(levelname)s: %(message)s'))
        logging.root.addHandler(handler)
    # Set logging level
    logging.root.setLevel(level)


def progress(count, total, prefix='', suffix='', length=60):
    """
    Show a progress bar
    :param count: current progress
    :param total: total progress
    :param prefix: prefix shown before the progress bar
    :param suffix: suffix shown after the progress bar
    :param length: length of the progress bar, default: 60
    :return: none
    """
    bar_len = length
    filled_len = int(round(bar_len * count / float(total)))
    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)
    sys.stdout.write('%s[%s] %s%s%s\r' % (prefix, bar, percents, '%', ' ' + suffix))
    sys.stdout.flush()


if __name__ == "__main__":
    setup_log(level=logging.DEBUG, path='./log/', tag='wrenchbox')
    logging.debug('This is a DEBUG message.')
    logging.info('This is an INFO message.')
    logging.warning('This is a WARNING message.')
    logging.error('This is an ERROR message.')
    logging.critical('This is an CRITICAL message.')
    for _i in range(100):
        progress(_i, 100)
        time.sleep(0.02)



import gzip
import logging
import os
import time
from datetime import datetime

from .logging import setup_log


class BufferedWriter:
    def __init__(self, path='.', prefix='', buffer=100000):
        """
        Buffered writer
        :param path: writing path where files will be stored as <timestamp>.dat.gz
        :param prefix: prefix for all files
        :param buffer: the buffer size
        """
        self.data = list()
        self.path = path
        self.prefix = prefix
        self.buffer = buffer

    def write(self, data):
        """
        Write data to file
        :param data: one line of data
        :return: none
        """
        self.data.append(data)
        if len(self.data) >= self.buffer:
            self.out()

    def out(self):
        """
        Immediately write all buffers to file
        :return: none
        """
        if len(self.data):
            batch = self.data[:self.buffer]
            self.data = self.data[self.buffer:]
            logging.info('Writing out {} records...'.format(len(batch)))
            if not os.path.exists(self.path):
                os.makedirs(self.path)
            with gzip.open(os.path.join(self.path, '{}{}.dat.gz'.format(self.prefix, int(datetime.now().timestamp()))), 'wb') as f:
                for i in batch:
                    f.write((str(i) + '\n').encode('utf-8'))

    def close(self):
        """
        Close the writer and write remaining buffer to file
        :return: none
        """
        self.out()


if __name__ == "__main__":
    setup_log()
    _w = BufferedWriter(buffer=50)
    for _i in range(100):
        _w.write(_i)
        time.sleep(0.02)
    _w.close()



from datetime import datetime

from dateutil import parser


class T:
    def __init__(self, t=None):
        if t is not None:
            if isinstance(t, int) or isinstance(t, float):
                self.t = datetime.fromtimestamp(t)
            elif isinstance(t, datetime):
                self.t = t
            elif isinstance(t, str):
                try:
                    self.t = datetime.fromtimestamp(int(t))
                except ValueError:
                    try:
                        self.t = datetime.fromtimestamp(float(t))
                    except ValueError:
                        self.t = parser.parse(t)
            else:
                raise ValueError
        else:
            self.t = datetime.now()

    def timestamp(self):
        """
        Get the timestamp of the datetime object
        :return: timestamp
        """
        return self.t.timestamp()

    def format(self, m: str = 'dt'):
        """
        Format the datetime object into various forms
        :param m: format, 'dt' or 'd' or 't'
        :return: the formatted string
        """
        if m == 'dt':
            return self.t.strftime('%Y-%m-%d %H:%M:%S')
        elif m == 'd':
            return self.t.strftime('%Y-%m-%d')
        elif m == 't':
            return self.t.strftime('%H:%M:%S')


if __name__ == "__main__":
    print(T().timestamp())
    print(T().format())
    print(T().format('d'))
    print(T().format('t'))



import argparse
import logging

import requests

from .logging import setup_log


class DingTalk:
    base_url = 'https://oapi.dingtalk.com/robot/'

    def __init__(self, access_token):
        self.access_token = access_token

    def send(self, msg, msg_type='text', title='ERROR'):
        m = {
            'msgtype': msg_type
        }
        if msg_type == 'text':
            m['text'] = {
                'content': msg
            }
        elif msg_type == 'markdown':
            m['markdown'] = {
                'title': title,
                'text': msg
            }
        else:
            raise ValueError('unknown msg type: {}'.format(msg_type))
        return requests.post('{}send?access_token={}'.format(self.base_url, self.access_token), json=m)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', action='store_true', default=False, help='show debug information')
    parser.add_argument('--type', '-p', type=str, required=True, help='alert type, must be "dingtalk"')
    parser.add_argument('--token', '-k', type=str, default=None, help='access token, if any')
    parser.add_argument('msg', type=str, help='the message')
    args, _ = parser.parse_known_args()
    setup_log(level=logging.DEBUG if args.debug else logging.INFO)
    print({'dingtalk': DingTalk}[args.type](args.token).send(args.msg).text)



import argparse
import logging
import re

from jieba import posseg

from .logging import setup_log


class S:
    def __init__(self, s):
        self.s = s
        self.words = None

    def v(self, chinese=False):
        if chinese:
            words = list()
            for w, a in posseg.cut(self.s):
                words.append((w, a))
        else:
            words = [(w, None) for w in re.sub("[^\w]", " ", self.s).split()]
        self.words = list()
        for w in set(words):
            self.words.append((w[0], w[1], words.count(w)))
        self.words = sorted(self.words, key=lambda x: x[2])
        self.words.reverse()
        return self


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', action='store_true', default=False, help='show debug information')
    parser.add_argument('--chinese', action='store_true', default=False, help='specify the input as Chinese')
    parser.add_argument('-p', type=str, default=None, help='part of speech to show, only works for Chinese')
    parser.add_argument('f', type=str, help='file path')
    args, _ = parser.parse_known_args()
    setup_log(level=logging.DEBUG if args.debug else logging.INFO)
    print([i for i in S(open(args.f, 'r').read()).v(chinese=args.chinese).words if i[1] in [i.strip() for i in args.p.split(',')]])



import random
import string
from math import pi


def digits(n):
    """
    Count the digits of a given float number
    :param n: the float number
    :return: number of digits
    """
    return len(str(n).split('.')[-1])


def random_chars(n):
    """
    Generate a random string from [a-zA-Z0-9]
    :param n: length of the string
    :return: the random string
    """
    return ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(n))


def random_letters(n):
    """
    Generate a random string from [a-zA-Z]
    :param n: length of the string
    :return: the random string
    """
    return ''.join(random.SystemRandom().choice(string.ascii_letters) for _ in range(n))


def random_numbers(n):
    """
    Generate a random string from [0-9]
    :param n: length of the string
    :return: the random string
    """
    return ''.join(random.SystemRandom().choice(string.digits) for _ in range(n))


if __name__ == "__main__":
    print(digits(pi))
    print(random_chars(3))
    print(random_letters(3))
    print(random_numbers(3))



import argparse
import logging
import time

from .logging import setup_log


class Snowflake:
    def __init__(self, twepoch):
        """
        Init a snowflake generator: https://developer.twitter.com/en/docs/twitter-ids
        :param twepoch: an initial timestamp when the database is created, e.g., 1483228800000
        """
        self.twepoch = twepoch
        self.worker_id_bits = 5
        self.data_center_id_bits = 5
        self.max_worker_id = -1 ^ (-1 << self.worker_id_bits)
        self.max_data_center_id = -1 ^ (-1 << self.data_center_id_bits)
        self.sequence_bits = 12
        self.worker_id_shift = self.sequence_bits
        self.data_center_id_shift = self.sequence_bits + self.worker_id_bits
        self.timestamp_left_shift = self.sequence_bits + self.worker_id_bits + self.data_center_id_bits
        self.sequence_mask = -1 ^ (-1 << self.sequence_bits)

    def timestamp(self, k):
        """
        Return the timestamp of a snowflake code
        :param k: the snowflake code
        :return: timestamp of the snowflake code
        """
        s = k >> 22
        s += self.twepoch
        return s / 1000

    def generate(self, worker_id, data_center_id=31, sleep=lambda x: time.sleep(x / 1000.0), count=1):
        """
        Generate a snowflake code
        :param worker_id: worker id, 0-31
        :param data_center_id: data center id, 0-31
        :param sleep: sleep time after one batch of codes are generated
        :param count: # of desired codes
        :return: a list of snowflake codes
        """
        assert 0 <= worker_id <= self.max_worker_id
        assert 0 <= data_center_id <= self.max_data_center_id
        last_timestamp = -1
        sequence = 0
        c = 0
        while c < count:
            timestamp = int(time.time() * 1000)
            if last_timestamp > timestamp:
                logging.warning("Clock is moving backwards, wait until: {}".format(last_timestamp))
                sleep(last_timestamp - timestamp)
                continue
            if last_timestamp == timestamp:
                sequence = (sequence + 1) & self.sequence_mask
                if sequence == 0:
                    logging.warning("Sequence overruns.")
                    sequence = -1 & self.sequence_mask
                    sleep(1)
                    continue
            else:
                sequence = 0
            last_timestamp = timestamp
            c += 1
            yield ((timestamp - self.twepoch) << self.timestamp_left_shift) | (data_center_id << self.data_center_id_shift) | (worker_id << self.worker_id_shift) | sequence


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', action='store_true', default=False, help='show debug information')
    parser.add_argument('--twepoch', type=int, default=1483228800000, help='twitter epoch, default: 1483228800000')
    parser.add_argument('-d', type=int, default=31, help='data center id, default: 31')
    parser.add_argument('-w', type=int, required=True, help='worker id, 0-31')
    parser.add_argument('n', type=int, help='# of results')
    args, _ = parser.parse_known_args()
    setup_log(level=logging.DEBUG if args.debug else logging.INFO)
    [print(i) for i in Snowflake(twepoch=args.twepoch).generate(args.w, data_center_id=args.d, count=args.n)]



import json


class Dict2StrSafe:
    # A class that correctly converts self.__dict__ to json string without errors
    def __str__(self):
        s = dict()
        for k, v in self.__dict__.items():
            if isinstance(v, list):
                s[k] = [json.loads(str(i)) if isinstance(i, Dict2StrSafe) else i for i in v]
            elif isinstance(v, Dict2StrSafe):
                s[k] = json.loads(str(v))
            else:
                s[k] = v
        return json.dumps(s, default=lambda i: str(i))


class Munch(Dict2StrSafe):
    def __init__(self, m: dict = None, **kwargs):
        """
        An alternative implementation of munch
        :param m: a dict
        """
        if m:
            for k, v in m.items():
                if isinstance(v, dict):
                    setattr(self, k, Munch(v))
                else:
                    setattr(self, k, v)
        for k, v in kwargs.items():
            if isinstance(v, dict):
                setattr(self, k, Munch(v))
            else:
                setattr(self, k, v)


if __name__ == '__main__':
    print(json.dumps(json.loads(str(Munch({'a': 1, 'b': {'a': 1, 'b': 2}}, a=3, c=1))), indent=2))



from os import path

from setuptools import setup, find_packages

with open(path.join(path.abspath(path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='wrenchbox',
    version='1.2.6',
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/valency/wrenchbox',
    author='Ye Ding',
    author_email='guiewy@gmail.com',
    description='Wrenchbox',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['toolbox', 'tools'],
    install_requires=[
        'django',
        'requests',
        'python-dateutil',
        'munch',
        'sqlalchemy',
        'ruamel.yaml',
        'jieba'
    ]
)



import argparse
import json
import logging
import uuid

import ruamel.yaml as yaml

from .logging import setup_log
from .object import Dict2StrSafe
from .object import Munch


class Configuration(Dict2StrSafe):
    def __init__(self, path: str, version: tuple = (0,)):
        """
        Create a configuration object
        :param path: the yaml or json file path
        :param version: accepted versions as a tuple
        """
        self.version = None
        self.worker = 0
        self.name = str(uuid.uuid4())
        with open(path) as f:
            data = f.read()
        if path.endswith('yml') or path.endswith('yaml'):
            data = yaml.safe_load(data)
        elif path.endswith('json'):
            data = json.loads(data)
        else:
            raise ValueError('config file must end with "yml", "yaml", or "json"')
        for k, v in data.items():
            if not k.startswith('_') and not k.startswith('.'):
                if isinstance(v, dict):
                    setattr(self, k, Munch(v))
                else:
                    setattr(self, k, v)
            else:
                logging.debug('Config key starts with "_" or "." will be ignored: {}'.format(k))
        if self.version is None:
            logging.warning('Configuration version is not set and may cause compatibility problems.')
        elif self.version not in version:
            raise ValueError('Version "{}" is not accepted, accepted versions: {}', self.version, ','.join(version))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', action='store_true', default=False, help='show debug information')
    parser.add_argument('config', type=str, help='config file')
    args, _ = parser.parse_known_args()
    setup_log(level=logging.DEBUG if args.debug else logging.INFO)
    print(json.dumps(json.loads(str(Configuration(args.config))), indent=2))



import argparse
import logging
from datetime import datetime

from munch import Munch
from sqlalchemy import create_engine, MetaData
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import Session

from .logging import setup_log


class Database:
    def __init__(self, url: str, schema: str = None, allow_empty: bool = False, tables=None):
        logging.debug('Connecting: %s', url)
        self.url = url
        self.schema = schema
        self.engine = create_engine(self.url, connect_args={'options': '-csearch_path={}'.format(self.schema)} if self.schema is not None else dict())
        self.session = Session(self.engine)
        try:
            self.base = automap_base()
            self.base.prepare(self.engine)
            MetaData().reflect(self.engine, only=tables)
        except Exception as exp:
            if allow_empty:
                logging.debug('Cannot map tables, an empty base will be created: %s', self.url)
                self.base = None
            else:
                raise exp
        logging.debug('Connection is established: %s', self.url)


class Table:
    def __init__(self, pool: Munch, database: str, table: str, alt_table=None):
        self.session = getattr(pool, database).session
        if alt_table is not None:
            logging.debug('Alternative table is provided: %s.%s', database, table)
            self.model = alt_table
        elif getattr(pool, database).base is not None:
            try:
                self.model = getattr(getattr(pool, database).base.classes, table)
            except AttributeError as exp:
                if table in getattr(pool, database).base.metadata.tables.keys():
                    logging.critical('Cannot find table (mainly caused by missing primary key): %s.%s', database, table)
                raise exp
        else:
            logging.critical('No database is provided for table: %s.%s', database, table)


class DatabaseHandler:
    def __init__(self, databases: dict, tables: list, alt_tables: dict = None):
        """
        Init a database handler
        :param databases: a dict of databases such like {'db':'url',...}
        :param tables: a list of tuples such like [('db','table'),...]
        """
        logging.debug('Init database handler...')
        self.config = Munch(
            databases=databases,
            tables=tables,
            alt_tables=alt_tables
        )
        self.o = Munch()
        self.db = Munch()
        if len(self.config.databases) and len(self.config.tables):
            self.connect(True)
        else:
            logging.debug('No database or table info are provided, an empty database handler will be created.')
        logging.debug('Initialization of database handler is completed.')

    def connect(self, reconnect: bool = False):
        if reconnect:
            self.o = Munch()
            self.db = Munch()
        for table in self.config.tables:
            if self.config.alt_tables is not None and table[1] in self.config.alt_tables:
                alt_table = self.config.alt_tables[table[1]]
            else:
                alt_table = None
            if getattr(self.db, table[0], None) is None:
                logging.debug('Creating database connection: %s', table[0])
                setattr(self.db, table[0], Database(
                    self.config.databases[table[0]],
                    schema='public' if self.config.databases[table[0]].startswith('postgresql') else None,
                    allow_empty=alt_table is not None,
                    tables=[i[1] for i in self.config.tables]
                ))
                logging.debug('Database connection is established: %s', table[0])
            setattr(self.o, table[1], Table(self.db, table[0], table[1], alt_table=alt_table))
            logging.debug('Table is synced: %s.%s', table[0], table[1])

    @staticmethod
    def commit(session, item):
        logging.debug(item.__dict__)
        session.add(item)
        try:
            session.commit()
        except Exception as exp:
            session.rollback()
            raise exp

    def handle(self, table: str, filters: dict, fields: dict, inserts: dict = None, replace: bool = True):
        """
        Perform insert or update query
        :param table: target table name
        :param filters: filters for identifying the unique record
        :param fields: fields for the insert or update
        :param inserts: fields for insert only despite of "fields"
        :param replace: no change will be applied to existing records if set to False
        :return:
        """
        c = getattr(self.o, table)
        s = c.session.query(c.model).filter_by(**filters)
        k = None
        if s.count() > 0:
            if s.count() == 1:
                k = s.first()
                if replace:
                    logging.debug('Targeting record of "%s" exists, will be replaced: %s', table, filters)
                    for key, value in fields.items():
                        setattr(k, key, value)
                    self.commit(c.session, k)
                else:
                    logging.debug('Targeting record of "%s" exists, will be kept: %s', table, filters)
            else:
                logging.error('Identifier of "%s" is not unique: %s', table, filters)
        else:
            k = c.model(**fields)
            if inserts:
                for key, value in inserts.items():
                    setattr(k, key, value)
            self.commit(c.session, k)
        return k


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', action='store_true', default=False, help='show debug information')
    parser.add_argument('--replace', action='store_true', default=False, help='replace existing records')
    parser.add_argument('--id', type=int, default=1, help='test id, default: 1')
    parser.add_argument('db', type=str, help='database url')
    parser.add_argument('table', type=str, help='table name, must have columns of "id" and "name"')
    args, _ = parser.parse_known_args()
    setup_log(level=logging.DEBUG if args.debug else logging.INFO)
    DatabaseHandler({
        'test': args.db
    }, [
        ('test', 'test')
    ]).handle(
        'test', {
            'id': args.id
        }, {
            'id': args.id,
            'name': str(datetime.now())
        }, replace=args.replace
    )
